@@ -1,6 +1,12 @@
-import datetime
 import json
+from datetime import date
+from datetime import datetime
 from time import mktime
+from typing import Iterable
+from typing import List
+from typing import Optional
+from typing import Tuple
+from typing import Union
 
 import click
 import humanize
@@ -9,20 +15,23 @@ import pytz
 from dateutil.tz import tzlocal
 from tabulate import tabulate
 
+from todoman.model import Todo
+from todoman.model import TodoList
 
-def rgb_to_ansi(colour):
+
+def rgb_to_ansi(colour: Optional[str]) -> Optional[str]:
     """
     Convert a string containing an RGB colour to ANSI escapes
     """
     if not colour or not colour.startswith("#"):
-        return
+        return None
 
     r, g, b = colour[1:3], colour[3:5], colour[5:7]
 
     if not len(r) == len(g) == len(b) == 2:
-        return
+        return None
 
-    return "\33[38;2;{!s};{!s};{!s}m".format(int(r, 16), int(g, 16), int(b, 16))
+    return f"\33[38;2;{int(r, 16)!s};{int(g, 16)!s};{int(b, 16)!s}m"
 
 
 class DefaultFormatter:
@@ -41,33 +50,31 @@ class DefaultFormatter:
         )
 
         self.tz = tz_override or tzlocal()
-        self.now = datetime.datetime.now().replace(tzinfo=self.tz)
+        self.now = datetime.now().replace(tzinfo=self.tz)
 
         self._parsedatetime_calendar = parsedatetime.Calendar(
             version=parsedatetime.VERSION_CONTEXT_STYLE,
         )
 
-    def simple_action(self, action, todo):
-        return '{} "{}"'.format(action, todo.summary)
+    def simple_action(self, action: str, todo: Todo) -> str:
+        return f'{action} "{todo.summary}"'
 
-    def compact(self, todo):
+    def compact(self, todo: Todo) -> str:
         return self.compact_multiple([todo])
 
-    def compact_multiple(self, todos, hide_list=False):
+    def compact_multiple(self, todos: Iterable[Todo], hide_list=False) -> str:
         table = []
         for todo in todos:
             completed = "X" if todo.is_completed else " "
             percent = todo.percent_complete or ""
             if percent:
-                percent = " ({}%)".format(percent)
+                percent = f" ({percent}%)"
             priority = self.format_priority_compact(todo.priority)
 
             due = self.format_datetime(todo.due)
-            now = (
-                self.now if isinstance(todo.due, datetime.datetime) else self.now.date()
-            )
+            now = self.now if isinstance(todo.due, datetime) else self.now.date()
             if todo.due and todo.due <= now and not todo.is_completed:
-                due = click.style(due, fg="red")
+                due = click.style(str(due), fg="red")
 
             recurring = "⟳" if todo.is_recurring else ""
 
@@ -77,6 +84,9 @@ class DefaultFormatter:
                     percent,
                 )
             else:
+                if not todo.list:
+                    raise ValueError("Cannot format todo without a list")
+
                 summary = "{} {}{}".format(
                     todo.summary,
                     self.format_database(todo.list),
@@ -86,40 +96,51 @@ class DefaultFormatter:
             table.append(
                 [
                     todo.id,
-                    "[{}]".format(completed),
+                    f"[{completed}]",
                     priority,
-                    "{} {}".format(due, recurring),
+                    f"{due} {recurring}",
                     summary,
                 ]
             )
 
         return tabulate(table, tablefmt="plain")
 
-    def _columnize_text(self, label, text):
-        """Display text, split text by line-endings, on multiple colums,"""
-        """do nothing if text is empty or None"""
+    def _columnize_text(
+        self,
+        label: str,
+        text: Optional[str],
+    ) -> List[Tuple[Optional[str], str]]:
+        """Display text, split text by line-endings, on multiple colums.
+
+        Do nothing if text is empty or None.
+        """
         lines = text.splitlines() if text else None
 
         return self._columnize_list(label, lines)
 
-    def _columnize_list(self, label, lst):
-        """Display list on multiple columns,"""
-        """do nothing if list is empty or None"""
+    def _columnize_list(
+        self,
+        label: str,
+        lst: Optional[List[str]],
+    ) -> List[Tuple[Optional[str], str]]:
+        """Display list on multiple columns.
 
-        rows = []
+        Do nothing if list is empty or None.
+        """
+
+        rows: List[Tuple[Optional[str], str]] = []
 
         if lst:
-            rows.append([label, lst[0]])
+            rows.append((label, lst[0]))
             for line in lst[1:]:
-                rows.append([None, line])
+                rows.append((None, line))
 
         return rows
 
-    def detailed(self, todo):
-        """
-        Returns a detailed representation of a task.
+    def detailed(self, todo: Todo) -> str:
+        """Returns a detailed representation of a task.
 
-        :param Todo todo: The todo component.
+        :param todo: The todo component.
         """
         extra_rows = []
         extra_rows += self._columnize_text("Description", todo.description)
@@ -131,15 +152,15 @@ class DefaultFormatter:
             )
         return self.compact(todo)
 
-    def format_datetime(self, dt):
+    def format_datetime(self, dt: Optional[date]) -> Union[str, int, None]:
         if not dt:
             return ""
-        elif isinstance(dt, datetime.datetime):
+        elif isinstance(dt, datetime):
             return dt.strftime(self.datetime_format)
-        elif isinstance(dt, datetime.date):
+        elif isinstance(dt, date):
             return dt.strftime(self.date_format)
 
-    def parse_priority(self, priority):
+    def parse_priority(self, priority: Optional[str]) -> Optional[int]:
         if priority is None or priority == "":
             return None
         if priority == "low":
@@ -153,7 +174,7 @@ class DefaultFormatter:
         else:
             raise ValueError("Priority has to be one of low, medium, high or none")
 
-    def format_priority(self, priority):
+    def format_priority(self, priority: Optional[int]) -> str:
         if not priority:
             return "none"
         elif 1 <= priority <= 4:
@@ -163,7 +184,7 @@ class DefaultFormatter:
         elif 6 <= priority <= 9:
             return "low"
 
-    def format_priority_compact(self, priority):
+    def format_priority_compact(self, priority: Optional[int]) -> str:
         if not priority:
             return ""
         elif 1 <= priority <= 4:
@@ -173,53 +194,53 @@ class DefaultFormatter:
         elif 6 <= priority <= 9:
             return "!"
 
-    def parse_datetime(self, dt):
+    def parse_datetime(self, dt: str) -> Optional[date]:
         if not dt:
             return None
 
         rv = self._parse_datetime_naive(dt)
-        return rv.replace(tzinfo=self.tz) if isinstance(rv, datetime.datetime) else rv
+        return rv.replace(tzinfo=self.tz) if isinstance(rv, datetime) else rv
 
-    def _parse_datetime_naive(self, dt):
+    def _parse_datetime_naive(self, dt: str) -> date:
         """Parse dt and returns a naive datetime or a date"""
         try:
-            return datetime.datetime.strptime(dt, self.datetime_format)
+            return datetime.strptime(dt, self.datetime_format)
         except ValueError:
             pass
 
         try:
-            return datetime.datetime.strptime(dt, self.date_format).date()
+            return datetime.strptime(dt, self.date_format).date()
         except ValueError:
             pass
 
         try:
-            return datetime.datetime.combine(
-                self.now.date(), datetime.datetime.strptime(dt, self.time_format).time()
+            return datetime.combine(
+                self.now.date(), datetime.strptime(dt, self.time_format).time()
             )
         except ValueError:
             pass
 
         rv, pd_ctx = self._parsedatetime_calendar.parse(dt)
         if not pd_ctx.hasDateOrTime:
-            raise ValueError("Time description not recognized: {}".format(dt))
-        return datetime.datetime.fromtimestamp(mktime(rv))
+            raise ValueError(f"Time description not recognized: {dt}")
+        return datetime.fromtimestamp(mktime(rv))
 
-    def format_database(self, database):
+    def format_database(self, database: TodoList):
         return "{}@{}".format(
             rgb_to_ansi(database.colour) or "", click.style(database.name)
         )
 
 
 class HumanizedFormatter(DefaultFormatter):
-    def format_datetime(self, dt):
+    def format_datetime(self, dt: Optional[date]) -> str:
         if not dt:
             return ""
 
-        if isinstance(dt, datetime.datetime):
+        if isinstance(dt, datetime):
             rv = humanize.naturaltime(self.now - dt)
             if " from now" in rv:
-                rv = "in {}".format(rv[:-9])
-        elif isinstance(dt, datetime.date):
+                rv = f"in {rv[:-9]}"
+        elif isinstance(dt, date):
             rv = humanize.naturaldate(dt)
 
         return rv
@@ -227,28 +248,30 @@ class HumanizedFormatter(DefaultFormatter):
 
 class PorcelainFormatter(DefaultFormatter):
     def _todo_as_dict(self, todo):
-        return dict(
-            completed=todo.is_completed,
-            due=self.format_datetime(todo.due),
-            id=todo.id,
-            list=todo.list.name,
-            percent=todo.percent_complete,
-            summary=todo.summary,
-            priority=todo.priority,
-            location=todo.location,
-        )
+        return {
+            "completed": todo.is_completed,
+            "due": self.format_datetime(todo.due),
+            "id": todo.id,
+            "list": todo.list.name,
+            "percent": todo.percent_complete,
+            "summary": todo.summary,
+            "priority": todo.priority,
+            "location": todo.location,
+            "description": todo.description,
+            "completed_at": self.format_datetime(todo.completed_at),
+        }
 
-    def compact(self, todo):
+    def compact(self, todo: Todo) -> str:
         return json.dumps(self._todo_as_dict(todo), indent=4, sort_keys=True)
 
-    def compact_multiple(self, todos, hide_list=False):
+    def compact_multiple(self, todos: Iterable[Todo], hide_list=False) -> str:
         data = [self._todo_as_dict(todo) for todo in todos]
         return json.dumps(data, indent=4, sort_keys=True)
 
-    def simple_action(self, action, todo):
+    def simple_action(self, action: str, todo: Todo) -> str:
         return self.compact(todo)
 
-    def parse_priority(self, priority):
+    def parse_priority(self, priority: Optional[str]) -> Optional[int]:
         if priority is None:
             return None
         try:
@@ -257,21 +280,23 @@ class PorcelainFormatter(DefaultFormatter):
             else:
                 raise ValueError("Priority has to be in the range 0-9")
         except ValueError as e:
-            raise click.BadParameter(e)
+            raise click.BadParameter(str(e))
 
-    def detailed(self, todo):
+    def detailed(self, todo: Todo) -> str:
         return self.compact(todo)
 
-    def format_datetime(self, date):
-        if date:
-            if not hasattr(date, "timestamp"):
-                date = datetime.datetime.fromordinal(date.toordinal())
-            return int(date.timestamp())
+    def format_datetime(self, value: Optional[date]) -> Optional[int]:
+        if value:
+            if not isinstance(value, datetime):
+                dt = datetime.fromordinal(value.toordinal())
+            else:
+                dt = value
+            return int(dt.timestamp())
         else:
             return None
 
     def parse_datetime(self, value):
         if value:
-            return datetime.datetime.fromtimestamp(value, tz=pytz.UTC)
+            return datetime.fromtimestamp(value, tz=pytz.UTC)
         else:
             return None
